import { isTrustedSender } from "../background/trust.js";

// methods holds a hash of the name of the message to the method to call.
const methods = {};
let context = {}; // the runtime context for the run function; blocker, etc
let trustedBaseUrl; // the base of trusted URLs (which originate from the extension)
let tabs;
let send;

export function setValidatedMethodBrowser({
  getURL,
  addListener,
  browserTabs,
  sendMessage,
}) {
  if (!getURL || !addListener || !browserTabs || !sendMessage) {
    return;
  }
  send = sendMessage;
  tabs = browserTabs;
  trustedBaseUrl = getURL("");
  // register a single top level handler for all messages
  addListener((msg, sender) => {
    const type = msg.type;
    const data = msg.data ? msg.data : msg;
    if (type && isTrustedSender(trustedBaseUrl, sender)) {
      const method = methods[type];
      if (method) {
        // TODO actually validate the data against the schema
        return method.options.run(data, context);
      }
    }
  });
}

export function setValidatedMethodContext(ctx) {
  context = ctx;
}

/**
 * Reload the browser tabs because of a change in ad blocker state.
 * This is a simple wrapper around the web extensions API so that we
 * can instrument this action later if we want to.
 */
function reloadTabs() {
  if (tabs.reload) {
    tabs.reload().then(() => {});
  }
}

/**
 * We needed to reload popover in Safari, so that we could
 * update popover according to the status of AdBlock.
 * We don't need to reload popup in Chrome because Chrome
 * reloads every time the popup for us.
 * Leaving the behavior centralized just in case Chrome changes.
 */
function closeAndReloadPopup() {
  window.close();
}

/**
 * ValidatedMethod provides a simple, centralized way to handle all
 * communication between background and popup.
 */
export class ValidatedMethod {
  /**
   * Create a ValidatedMethod that can be called.
   *
   * @param name {string} name of the validated method - must be unique
   * @param [mixins] {[]function} array of optional mixins that will modify the validated method behavior (logging, auth)
   * @param [schema] {object} the schema to validate message data against
   * @param run {function} the function to run with validated parameters
   */
  constructor({ name, run, mixins, schema }) {
    this.options = {
      name,
      mixins,
      schema,
      run,
    };
    methods[name] = this;
  }

  /**
   * Call the validated method (from the popup etc.) with the provided data.
   * Returns a promise that resolves to any response from the background page.
   *
   * @param [data] {Object} optional data to send along with the message.
   * @returns {Promise<Object|void>} Promise that resolves to the response to the message (if any)
   */
  async call(data) {
    if (send) {
      data = data || {};
      let tabId;
      if (
        document.location.search &&
        document.location.search.indexOf("tabId") > 0
      ) {
        const params = new URLSearchParams(document.location.search);
        tabId = params.get("tabId");
        if (tabId === "error") {
          // allows testing of the error handling logic
          throw new Error("anError");
        }
        let tab = tabs.get(tabId);
        data.url = tab.url;
        data.tabId = tabId;
        return send({
          type: this.options.name,
          data,
        });
      }
      const activeTabs = await tabs.query({
        active: true,
        lastFocusedWindow: true,
      });
      if (activeTabs.length > 0) {
        const tab = activeTabs[0];
        data.url = tab.url;
        data.tabId = tab.id;
        return send({
          type: this.options.name,
          data,
        });
      }
    } else {
      return alert(`sent message ${this.options.name}`);
    }
  }

  /**
   * Call the validated method (from the popup etc.) with the provided data.
   * After the validated method is finished, then close and reload the popup.
   * Returns a promise that resolves when the window has been closed.
   *
   * @param [data] {Object} optional data to send along with the message
   * @returns {Promise<void>} Promise that resolves when the popup has been reloaded
   */
  async callAndReload(data) {
    return this.call(data).then(() => {
      reloadTabs();
      closeAndReloadPopup();
    });
  }
}
